@@ -2,18 +2,16 @@
 
 import io
 import json
-import threading
-import time
 import os
 import signal
-from collections import namedtuple
+import time
 
+from sakee import addoninfo
+from sakee.colors import Colors
+from sakee.internalplayer import KodiInteralPlayer
+from sakee.stub import KodiStub
 from xbmcgui import ListItem
 
-from colors import Colors
-from stub import KodiStub
-
-
 DRIVE_NOT_READY = 1
 ENGLISH_NAME = 2
 ISO_639_1 = 0
@@ -40,15 +38,6 @@ TRAY_CLOSED_NO_MEDIA = 64
 TRAY_OPEN = 16
 
 
-# Custom AddonData type
-AddonData = namedtuple('AddonData', [
-    'kodi_home_path',       # data path (either portable of user path)
-    'add_on_id',            # the add-on id
-    'add_on_path',          # the full path to the add-on
-    'kodi_profile_path'     # the full path to the add-on profile folder
-])
-
-
 # noinspection PyPep8Naming
 class Monitor(KodiStub):
     def __init__(self):
@@ -86,13 +75,18 @@ class Monitor(KodiStub):
 
         """
 
-        time.sleep(seconds)
-        return self.__abort
+        deadline = time.time() + seconds
+        while time.time() < deadline:
+            if self.__abort:
+                return True
+            time.sleep(0.1)  # Sleep 100ms
+
+        return False
 
 
 # noinspection PyPep8Naming
 class Keyboard(KodiStub):
-    def __init__(self, line, heading, hidden):
+    def __init__(self, line="", heading="", hidden=False):
         """
 
         :param str line:        The default line to show?
@@ -120,7 +114,9 @@ class Keyboard(KodiStub):
 
         if not self.is_interactive:
             self.__confirmed = True
-            return "1234"
+            keyboard = self.get_keyboard_stub()
+            self.__input = keyboard.get_next_input()
+            return
 
         KodiStub.print_heading(self.__heading)
         try:
@@ -227,7 +223,7 @@ class PlayList(KodiStub):
 
     def add(self, url, listitem=None, index=None):
         """`Adds a new file to the playlist.
-        
+
         :param str url:                 Filename or url to add.
         :param ListItem|None listitem:  Used with setInfo() to set different infolabels.
         :param int|None index:          Position to add playlist item. (default=end)
@@ -247,22 +243,195 @@ class PlayList(KodiStub):
 
 # noinspection PyPep8Naming,PyArgumentList
 class Player(KodiStub):
-    def __init__(self, *args, **kwargs):
+    def __init__(self):
         super(Player, self).__init__()
 
-        self.Init(*args, **kwargs)
-        self.play_started = False
+        # register this player with the main Kodi player
+        KodiInteralPlayer.instance().register_player(self)
 
-        def play_start():
-            self.play_started = True
-            self.onAVStarted()
+    def __del__(self):
+        # unregister
+        KodiInteralPlayer.instance().unregister_player(self)
 
-        timer = threading.Timer(2.0, play_start)
-        timer.start()
+    def getAvailableAudioStreams(self):  # NOSONAR
+        """ Returns the available audio stream names.
+
+        :return: The available audio streams.
+        :rtype: List[str]
+
+        """
+
+        return []  # Not implemented
+
+    def getAvailableSubtitleStreams(self):  # NOSONAR
+        """ Returns the available subtitle stream names.
+
+        :return: The available subtitle streams.
+        :rtype: List[str]
+
+        """
+
+        return []  # Not implemented
+
+    def getAvailableVideoStreams(self):  # NOSONAR
+        """ Returns the available video stream names.
+
+        :return: The available video streams.
+        :rtype: List[str]
+
+        """
+
+        return []  # Not implemented
+
+    def getMusicInfoTag(self):
+        """ Return the music info tag.
+
+        :return: Returns the MusicInfoTag of the current playing 'Song'.
+        :rtype: MusicInfoTag|None
+
+        """
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            raise Exception('Player is not playing a file.')
+
+        return None  # Not implemented
+
+    def getPlayingFile(self):  # NOSONAR
+        """ Returns the current playing file as a string.
+
+        :return: The filename of the playing item.
+        :rtype: str
+
+        """
+
+        return KodiInteralPlayer.instance().file
+
+    def getRadioRDSInfoTag(self):
+        """ Return the Radio RDS info tag.
+
+        :return: Returns the RadioRDSInfoTag of the current playing 'Radio Song if. present'.
+        :rtype: RadioRDSInfoTag|None
+        """
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            raise Exception('Player is not playing a file.')
+
+        return None  # Not implemented
+
+    def getSubtitles(self):
+        """ Return the subtitle stream name.
+
+        :return: Stream name
+        :rtype: str
+
+        """
+
+        return ''  # Not implemented
+
+    def getTime(self):
+        """ Return the current playing time.
+
+        :return: Returns the current time of the current playing media as fractional seconds.
+        :rtype: int
+        """
+
+        return KodiInteralPlayer.instance().current_time
+
+    def getTotalTime(self):
+        """ Return the total playing time.
+
+        :return: Returns the total time of the current playing media in seconds. This is only accurate to the full second.
+        :rtype: int
+
+        """
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            raise Exception('Player is not playing a file.')
+
+        return KodiInteralPlayer.instance().total_time
+
+    def getVideoInfoTag(self):
+        """ Return the video info tag.
+
+        :return: The VideoInfoTag of the current playing Movie.
+        :rtype: VideoInfoTag|None
+
+        """
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            raise Exception('Player is not playing a file.')
+
+        return None  # Not implemented
+
+    def isExternalPlayer(self):
+        """ Check for external player.
+
+        :return: True if kodi is playing using an external player.
+        :rtype: bool
+
+        """
+
+        return False
+
+    def isPlaying(self):  # NOSONAR
+        """ Check Kodi is playing something.
+
+        :return: True if Kodi is playing a file.
+        :rtype: bool
+
+        """
+
+        return KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_PLAYING
+
+    def isPlayingAudio(self):  # NOSONAR
+        """ Check for playing audio.
+
+        :return: True if Kodi is playing an audio file.
+        :rtype: bool
+
+        """
+
+        return KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_PLAYING
+
+    def isPlayingRDS(self):  # NOSONAR
+        """ Check for playing radio data system (RDS).
+
+        :return: True if kodi is playing a radio data system (RDS).
+        :rtype: bool
+
+        """
+
+        return KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_PLAYING
+
+    def isPlayingVideo(self):  # NOSONAR
+        """ Check for playing video.
+
+        :return: True if Kodi is playing a video.
+        :rtype: bool
+
+        """
+
+        return KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_PLAYING
+
+    def pause(self):  # NOSONAR
+        """ Pause or resume playing if already paused. """
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            return
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_PLAYING:
+            KodiInteralPlayer.instance().status = KodiInteralPlayer.STATUS_PAUSED
+            self.onPlayBackPaused()
+            return
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_PAUSED:
+            KodiInteralPlayer.instance().status = KodiInteralPlayer.STATUS_PLAYING
+            self.onPlayBackResumed()
+            return
 
     # noinspection PyUnusedLocal
     def play(self, item=None, listitem=None, windowed=False, startpos=-1):
-        """ Play a item.
+        """ Play an item.
 
         :param str|None item:            Filename, url or playlist
         :param ListItem|None listitem:   Used with setInfo() to set different infolabels.
@@ -275,29 +444,227 @@ class Player(KodiStub):
         Once you use a keyword, all following arguments require the keyword.
 
         """
+        # Stop playing the current file (if any)
+        self.stop()
+        KodiInteralPlayer.instance().play(item)
 
-        time.sleep(1)
-        self.onAVStarted()
+    def playnext(self):  # NOSONAR
+        """ Play next item in playlist."""
+
+        pass  # Not implemented
+
+    def playprevious(self):  # NOSONAR
+        """ Play previous item in playlist."""
+
+        pass  # Not implemented
+
+    def playselected(self, selected):  # NOSONAR
+        """ Set Audio Stream.
+
+        :param int selected:            Item to select
 
-    def isPlaying(self):  # NOSONAR
-        # type: () -> bool
         """
-        Check Kodi is playing something.
 
-        :return: True if Kodi is playing a file.
+        pass  # Not implemented
+
+    def seekTime(self, seekTime):  # NOSONAR
+        """ Seek time.
+
+        Seeks the specified amount of time as fractional seconds.
+        The time specified is relative to the beginning of the currently. playing media file.
+
+        :param int seekTime:            Time to seek as fractional seconds
+
         """
-        return self.play_started
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            raise Exception('Player is not playing a file.')
+
+        KodiInteralPlayer.instance().current_time = seekTime
+        self.onPlayBackSeek(seekTime, 0)
+
+    def setAudioStream(self, stream):  # NOSONAR
+        """ Set Audio Stream.
+
+        :param int stream:              Audio stream to select for play
+
+        """
+
+        pass  # Not implemented
+
+    def setSubtitles(self, subtitleFile):  # NOSONAR
+        """ Set subtitle file and enable subtitles.
+
+        :param str subtitleFile:        File to use as source of subtitles
+
+        """
+
+        pass  # Not implemented
+
+    def setSubtitleStream(self, stream):  # NOSONAR
+        """ Set Subtitle Stream.
+
+        :param int stream:              Subtitle stream to select for play
+
+        """
+
+        pass  # Not implemented
+
+    def setVideoStream(self, stream):  # NOSONAR
+        """ Set Video Stream.
+
+        :param int stream:              Video stream to select for play
+
+        """
+
+        pass  # Not implemented
+
+    def showSubtitles(self, visible):  # NOSONAR
+        """ Enable / disable subtitles.
+
+        :param bool visible:            True for visible subtitles.
+
+        """
+
+        pass  # Not implemented
+
+    def stop(self):  # NOSONAR
+        """ Stop playing."""
+
+        KodiInteralPlayer.instance().status = KodiInteralPlayer.STATUS_STOPPED
+        KodiInteralPlayer.instance().current_time = 0
+        KodiInteralPlayer.instance().total_time = 0
+        KodiInteralPlayer.instance().file = None
+        KodiInteralPlayer.instance().stop_playback()
+
+    # noinspection PyUnusedLocal
+    def updateInfoTag(self, item):  # NOSONAR
+        """ Update info labels for currently playing item.
+
+        :param ListItem item:           ListItem with new info
+
+        """
+
+        if KodiInteralPlayer.instance().status == KodiInteralPlayer.STATUS_STOPPED:
+            raise Exception('Player is not playing a file.')
+
+        # Not implemented
+
+    def onAVChange(self):  # NOSONAR
+        """ onAVChange method.
+        Will be called when Kodi has a video, audio or subtitle stream. Also happens when the stream changes.
+
+        """
+
+        self.print_line('Invoked onAVChange()', verbose=True)
 
     def onAVStarted(self):  # NOSONAR
-        # type: () -> None
-        """
-        onAVStarted method.
+        """ onAVStarted method.
 
         Will be called when Kodi has a video or audiostream.
 
-        New function added.
         """
-        pass
+
+        self.print_line('Invoked onAVStarted()', verbose=True)
+
+    def onPlaybackEnded(self):  # NOSONAR
+        """ onPlaybackEnded method.
+
+        Will be called when Kodi stops playing a file.
+
+        """
+        self.print_line('Invoked onPlaybackEnded()', verbose=True)
+
+    def onPlayBackError(self):  # NOSONAR
+        """ onPlayBackError method.
+
+        Will be called when playback stops due to an error.
+
+        """
+
+        self.print_line('Invoked onPlayBackError()', verbose=True)
+
+    def onPlayBackPaused(self):  # NOSONAR
+        """ onPlayBackPaused method.
+
+        Will be called when user pauses a playing file.
+
+        """
+
+        self.print_line('Invoked onPlayBackPaused()', verbose=True)
+
+    def onPlayBackResumed(self):  # NOSONAR
+        """ onPlayBackResumed method.
+
+        Will be called when user resumes a paused file.
+
+        """
+
+        self.print_line('Invoked onPlayBackResumed()', verbose=True)
+
+    def onPlayBackSeek(self, time, seekOffset):  # NOSONAR
+        """ onPlayBackSeek method.
+        Will be called when user seeks to a time.
+
+        :param int time:            Time to seek to
+        :param int seekOffset:      ?
+
+        """
+        self.print_line('Invoked onPlayBackSeek(%d, %d)' % (time, seekOffset), verbose=True)
+
+    def onPlayBackSeekChapter(self, chapter):  # NOSONAR
+        """ onPlayBackSeekChapter method.
+
+        Will be called when user performs a chapter seek.
+
+        :param int chapter:         Chapter to seek to
+
+        """
+
+        self.print_line('Invoked onPlayBackSeekChapter(%d)' % chapter, verbose=True)
+
+    def onPlayBackSpeedChanged(self, speed):  # NOSONAR
+        """ onPlayBackSpeedChanged method.
+
+        Will be called when players speed changes (eg. user FF/RW).
+
+        Negative speed means player is rewinding, 1 is normal playback speed.
+
+        :param int speed:           Current speed of player
+
+        """
+
+        self.print_line('Invoked onPlayBackSpeedChanged(%d)' % speed, verbose=True)
+
+    def onPlayBackStarted(self):  # NOSONAR
+        """ onPlayBackStarted method.
+
+        Will be called when Kodi player starts. Video or audio might not be available at this point.
+
+        Use onAVStarted() instead if you need to detect if Kodi is actually playing a media file
+        (i.e, if a stream is available).
+
+        """
+
+        self.print_line('Invoked onPlayBackStarted()', verbose=True)
+
+    def onPlayBackStopped(self):  # NOSONAR
+        """ onPlayBackStopped method.
+
+        Will be called when user stops Kodi playing a file.
+
+        """
+
+        self.print_line('Invoked onPlayBackStopped()', verbose=True)
+
+    def onQueueNextItem(self):  # NOSONAR
+        """ onQueueNextItem method.
+
+        Will be called when user queues the next item.
+
+        """
+
+        self.print_line('Invoked onQueueNextItem()', verbose=True)
 
 
 # noinspection PyPep8Naming
@@ -310,22 +677,38 @@ def executeJSONRPC(jsonrpccommand):  # NOSONAR
     :rtype: str
 
     See https://codedocs.xyz/xbmc/xbmc/namespace_j_s_o_n_r_p_c.html
-
     """
+    from sakee.sakejsonrpc import JsonRpcApi
 
     json_data = json.loads(jsonrpccommand)
-    json_method = json_data["method"]
+    try:
+        # Implement some methods for real
+        return json.dumps(JsonRpcApi().handle(json_data))
+
+    except NotImplementedError:
+        # Fallback to stubs
+        pass
+
     json_responses = os.environ.get("KODI_STUB_RPC_RESPONSES")
     if not json_responses:
-        raise ValueError("Could not find JSON Response folder. Use the environment variable KODI_STUB_RPC_RESPONSES to set one")
+        raise ValueError(
+            "Could not find JSON Response folder. Use the environment variable KODI_STUB_RPC_RESPONSES to set one")
 
-    json_response = os.path.join(json_responses, "{}.json".format(json_method.lower()))
+    path = "{}.json".format(os.path.join(json_responses, json_data["method"].lower()))
+    if os.path.isfile(path):
+        with io.open(path, mode='r', encoding='utf-8') as fd:
+            stub_content = json.loads(fd.read())
+            if isinstance(stub_content, dict):
+                return json.dumps(stub_content)
+            try:
+                return json.dumps(next(stub.get('response')
+                                       for stub in stub_content
+                                       if json_data.get('params')
+                                       and stub.get('request', {}).get('method') == json_data.get('method')
+                                       and stub.get('request', {}).get('params') == json_data.get('params')))
+            except StopIteration:
+                pass
 
-    if os.path.isfile(json_response):
-        with io.open(json_response, mode='r', encoding='utf-8') as fd:
-            return fd.read()
-
-    # {"error":{"code":-32602,"message":"Invalid params."},"id":1,"jsonrpc":"2.0"}
     return '{"id":1,"jsonrpc":"2.0","result":"OK"}'
 
 
@@ -350,33 +733,9 @@ def translatePath(path):  # NOSONAR
 
     """
 
-    if path.startswith("special://profile/"):
-        sub_path = path.replace("special://profile/", "")
-        if not __add_on_info.kodi_profile_path:
-            raise ValueError("Missing kodi_profile_path data")
-
-        return_path = os.path.join(__add_on_info.kodi_profile_path, sub_path.replace("/", os.sep))
-        if not os.path.exists(return_path):
-            raise ValueError("Invalid path in profile specified: %s" % (path,))
-
-    elif path.startswith("special://home/"):
-        sub_path = path.replace("special://home/", "")
-        if not __add_on_info.kodi_home_path:
-            raise ValueError("Missing __kodi_home_path data")
-
-        return_path = os.path.join(__add_on_info.kodi_home_path, sub_path.replace("/", os.sep))
-        if not os.path.isdir(return_path):
-            raise ValueError("Invalid Kodi Home path: %s" % (path,))
-
-    elif os.path.isabs(path):
-        return path
-
-    else:
-        raise ValueError("Invalid special path: %s" % (path,))
-
-    actual_path = os.path.abspath(return_path)
-    KodiStub.print_line("Mapped '{0}' -> '{1}'".format(path, actual_path), color=Colors.Blue)
-    return actual_path
+    KodiStub.print_line("Warning: 'xbmc.translatePath' was moved to xbmcvfs in Kodi Matrix (API version 19.0)", color=Colors.Red)
+    import xbmcvfs
+    return xbmcvfs.translatePath(path)
 
 
 def executebuiltin(function):
@@ -398,7 +757,7 @@ def getCondVisibility(condition):  # NOSONAR
     :param str condition:   condition to check
 
     :return: True (1) or False (0) as a bool
-    :rtype: bool
+    :rtype: int
 
     List of Conditions - http://kodi.wiki/view/List_of_Boolean_Conditions
 
@@ -427,7 +786,7 @@ def getCondVisibility(condition):  # NOSONAR
         KodiStub.print_line("Missing condition: {}".format(condition), color=Colors.Yellow)
 
     KodiStub.print_line("Condition: {0}={1}".format(condition, result), color=Colors.Blue, verbose=True)
-    return result
+    return 1 if result else 0
 
 
 # noinspection PyPep8Naming
@@ -442,18 +801,18 @@ def getInfoLabel(infoTag):  # NOSONAR
     List of InfoTags - http://kodi.wiki/view/InfoLabels
     """
 
-    if infoTag == "system.buildversion":
-        return "18.1 Git:20160424-c327c53"
-    
+    if infoTag.lower() == "system.buildversion":
+        return "19.0 Git:20200626-xxxxxxxxxx"
+
     return "InfoLabel:{}".format(infoTag)
 
 
 # noinspection PyPep8Naming,PyShadowingBuiltins,PyUnusedLocal
-def getLanguage(format=ISO_639_1, region=None):  # NOSONAR
+def getLanguage(format=ENGLISH_NAME, region=False):  # NOSONAR
     """ Get the active language.
 
     :param int format:          Format of the returned language string (see table)
-    :param str|None region:     append the region delimited by "-" of the language
+    :param bool region:         append the region delimited by "-" of the language
                                 (setting) to the returned language string
 
     :return: The active language as a string
@@ -471,9 +830,9 @@ def getLanguage(format=ISO_639_1, region=None):  # NOSONAR
     """
 
     if format == ISO_639_1:
-        return "en"
+        return "en" if not region else "en-gb"
     if format == ISO_639_2:
-        return "eng"
+        return "eng" if not region else "eng-gbr"
 
     return "English"
 
@@ -508,80 +867,4 @@ def log(msg, level=0):
         print(msg)
 
 
-def get_add_on_info_from_calling_script(add_on_id=None, print_info=False):
-    if add_on_id is not None:
-        # Always print details for specific add-ons
-        print_info = True
-
-    # What is the Kodi Home path?
-    assert os.path.isfile("addon.xml"), "Working directory outside add-on path: {}".format(os.getcwd())
-    calling_add_on_path = os.getcwd()
-
-    # Get the generic Kodi paths
-    kodi_home_path = os.environ.get("KODI_HOME")
-    if kodi_home_path:
-        # We only need the ID from the path.
-        _, path_add_on_id = calling_add_on_path.rsplit(os.sep, 1)
-    else:
-        # determine it based on the running path
-        kodi_home_path, addon_name, path_add_on_id = os.getcwd().rsplit(os.sep, 2)
-
-    kodi_home_path = os.path.abspath(kodi_home_path)
-    assert os.path.isdir(kodi_home_path), \
-        "Kodi home path (special://home) does not exist: '{}'".format(kodi_home_path)
-
-    # Find the very first script called to determine the add-on ID if it was not specified
-    add_on_id = add_on_id or path_add_on_id
-
-    # Active add-on path
-    if add_on_id is None or add_on_id == path_add_on_id:
-        # We should use the data from the current calling add-on
-        add_on_path = calling_add_on_path
-    else:
-        # We should set it to the requested add-on based on the given add-on ID
-        add_on_path = os.path.join(kodi_home_path, "addons", add_on_id)
-        if not os.path.isdir(add_on_path) and "portable_data" in calling_add_on_path:
-            add_on_path = os.path.abspath(os.path.join(kodi_home_path, "..", "addons", add_on_id))
-    assert os.path.isdir(add_on_path), \
-        "Invalid add-on dir for add-on '{}': {}".format(add_on_id, add_on_path)
-
-    # The active profile path
-    if "KODI_PROFILE" in os.environ:
-        add_on_profile_path = os.path.abspath(os.environ["KODI_PROFILE"])
-    else:
-        add_on_profile_path = os.path.join(kodi_home_path, "userdata")
-
-    add_on_profile_path = os.path.abspath(add_on_profile_path)
-    assert os.path.isdir(add_on_profile_path), \
-        "Invalid Kodi master profile dir (special://masterprofile): {}".format(add_on_profile_path)
-
-    if "KODI_ACTIVE_PROFILE" in os.environ:
-        add_on_profile_path = os.path.join(
-            add_on_profile_path, "profiles", os.environ["KODI_ACTIVE_PROFILE"])
-
-        assert os.path.isdir(add_on_profile_path), \
-            "Invalid Kodi profile dir (special://profile): {}".format(add_on_profile_path)
-
-    a = AddonData(
-        kodi_home_path=kodi_home_path,
-        add_on_id=add_on_id,
-        add_on_path=add_on_path,
-        kodi_profile_path=add_on_profile_path
-    )
-
-    if not print_info:
-        return a
-
-    KodiStub.print_line(
-        "Found Add-on info: \n"
-        "- Kodi Home (special://home):       {} \n"
-        "- Add-on ID:                        {} \n"
-        "- Add-on Path:                      {} \n"
-        "- Kodi Profile (special://profile): {} \n"
-        .format(a.kodi_home_path, a.kodi_profile_path, a.add_on_id, a.add_on_path),
-        color=Colors.Blue
-    )
-    return a
-
-
-__add_on_info = get_add_on_info_from_calling_script(print_info=True)
+__add_on_info = addoninfo.get_add_on_info_from_calling_script(print_info=True)
